@@ -15,13 +15,15 @@ from fnmatch import fnmatchcase
 from subprocess import list2cmdline
 from threading import Thread
 
-import pkg_resources
 import pluggy
 import py
 import toml
+from packaging import requirements
+from packaging.utils import canonicalize_name
 
 import tox
 from tox.constants import INFO
+from tox.exception import MissingDependency
 from tox.interpreters import Interpreters, NoInterpreterInfo
 from tox.reporter import (
     REPORTER_TIMESTAMP_ON_ENV,
@@ -31,23 +33,25 @@ from tox.reporter import (
     verbosity1,
 )
 from tox.util.path import ensure_empty_dir
+from tox.util.stdlib import importlib_metadata
 
-from .parallel import ENV_VAR_KEY as PARALLEL_ENV_VAR_KEY
+from .parallel import ENV_VAR_KEY_PRIVATE as PARALLEL_ENV_VAR_KEY_PRIVATE
+from .parallel import ENV_VAR_KEY_PUBLIC as PARALLEL_ENV_VAR_KEY_PUBLIC
 from .parallel import add_parallel_config, add_parallel_flags
 from .reporter import add_verbosity_commands
 
+try:
+    from shlex import quote as shlex_quote
+except ImportError:
+    from pipes import quote as shlex_quote
+
+
 hookimpl = tox.hookimpl
-"""DEPRECATED - REMOVE - this is left for compatibility with plugins importing this from here.
+"""DEPRECATED - REMOVE - left for compatibility with plugins importing from here.
 
-Instead create a hookimpl in your code with:
-
-    import pluggy
-    hookimpl = pluggy.HookimplMarker("tox")
+Import hookimpl directly from tox instead.
 """
 
-default_factors = tox.PYTHON.DEFAULT_FACTORS
-"""DEPRECATED MOVE - please update to new location."""
-
 WITHIN_PROVISION = os.environ.get(str("TOX_PROVISION")) == "1"
 
 
@@ -122,7 +126,7 @@ class Parser:
     def parse_cli(self, args, strict=False):
         args, argv = self.argparser.parse_known_args(args)
         if argv and (strict or WITHIN_PROVISION):
-            self.argparser.error("unrecognized arguments: %s".format(" ".join(argv)))
+            self.argparser.error("unrecognized arguments: {}".format(" ".join(argv)))
         return args
 
     def _format_help(self):
@@ -192,10 +196,10 @@ class DepOption:
     @classmethod
     def _is_same_dep(cls, dep1, dep2):
         """Definitions are the same if they refer to the same package, even if versions differ."""
-        dep1_name = pkg_resources.Requirement.parse(dep1).project_name
+        dep1_name = canonicalize_name(requirements.Requirement(dep1).name)
         try:
-            dep2_name = pkg_resources.Requirement.parse(dep2).project_name
-        except pkg_resources.RequirementParseError:
+            dep2_name = canonicalize_name(requirements.Requirement(dep2).name)
+        except requirements.InvalidRequirement:
             # we couldn't parse a version, probably a URL
             return False
         return dep1_name == dep2_name
@@ -377,14 +381,9 @@ class SetenvDict(object):
 @tox.hookimpl
 def tox_addoption(parser):
     parser.add_argument(
-        "--version",
-        action="store_true",
-        dest="version",
-        help="report version information to stdout.",
-    )
-    parser.add_argument(
-        "-h", "--help", action="store_true", dest="help", help="show help about options"
+        "--version", action="store_true", help="report version information to stdout."
     )
+    parser.add_argument("-h", "--help", action="store_true", help="show help about options")
     parser.add_argument(
         "--help-ini", "--hi", action="store_true", dest="helpini", help="show help about ini-names"
     )
@@ -399,22 +398,16 @@ def tox_addoption(parser):
         "-l",
         "--listenvs",
         action="store_true",
-        dest="listenvs",
         help="show list of test environments (with description if verbose)",
     )
     parser.add_argument(
         "-a",
         "--listenvs-all",
         action="store_true",
-        dest="listenvs_all",
         help="show list of all defined environments (with description if verbose)",
     )
     parser.add_argument(
-        "-c",
-        action="store",
-        default=None,
-        dest="configfile",
-        help="config file name or directory with 'tox.ini' file.",
+        "-c", dest="configfile", help="config file name or directory with 'tox.ini' file."
     )
     parser.add_argument(
         "-e",
@@ -424,13 +417,16 @@ def tox_addoption(parser):
         help="work against specified environments (ALL selects all).",
     )
     parser.add_argument(
-        "--notest", action="store_true", dest="notest", help="skip invoking test commands."
+        "--devenv",
+        metavar="ENVDIR",
+        help=(
+            "sets up a development environment at ENVDIR based on the env's tox "
+            "configuration specified by `-e` (-e defaults to py)."
+        ),
     )
+    parser.add_argument("--notest", action="store_true", help="skip invoking test commands.")
     parser.add_argument(
-        "--sdistonly",
-        action="store_true",
-        dest="sdistonly",
-        help="only perform the sdist packaging activity.",
+        "--sdistonly", action="store_true", help="only perform the sdist packaging activity."
     )
     add_parallel_flags(parser)
     parser.add_argument(
@@ -442,15 +438,12 @@ def tox_addoption(parser):
     )
     parser.add_argument(
         "--installpkg",
-        action="store",
-        default=None,
         metavar="PATH",
         help="use specified package for installation into venv, instead of creating an sdist.",
     )
     parser.add_argument(
         "--develop",
         action="store_true",
-        dest="develop",
         help="install package in the venv using 'setup.py develop' via 'pip -e .'",
     )
     parser.add_argument(
@@ -465,21 +458,15 @@ def tox_addoption(parser):
     parser.add_argument(
         "--pre",
         action="store_true",
-        dest="pre",
         help="install pre-releases and development versions of dependencies. "
         "This will pass the --pre option to install_command "
         "(pip by default).",
     )
     parser.add_argument(
-        "-r",
-        "--recreate",
-        action="store_true",
-        dest="recreate",
-        help="force recreation of virtual environments",
+        "-r", "--recreate", action="store_true", help="force recreation of virtual environments"
     )
     parser.add_argument(
         "--result-json",
-        action="store",
         dest="resultjson",
         metavar="PATH",
         help="write a json file with detailed information "
@@ -489,9 +476,7 @@ def tox_addoption(parser):
     # We choose 1 to 4294967295 because it is the range of PYTHONHASHSEED.
     parser.add_argument(
         "--hashseed",
-        action="store",
         metavar="SEED",
-        default=None,
         help="set PYTHONHASHSEED to SEED before running commands.  "
         "Defaults to a random integer in the range [1, 4294967295] "
         "([1, 1024] on Windows). "
@@ -501,7 +486,6 @@ def tox_addoption(parser):
         "--force-dep",
         action="append",
         metavar="REQ",
-        default=None,
         help="Forces a certain version of one of the dependencies "
         "when configuring the virtual environment. REQ Examples "
         "'pytest<2.7' or 'django>=1.6'.",
@@ -516,25 +500,25 @@ def tox_addoption(parser):
     )
 
     cli_skip_missing_interpreter(parser)
-    parser.add_argument(
-        "--workdir",
-        action="store",
-        dest="workdir",
-        metavar="PATH",
-        default=None,
-        help="tox working directory",
-    )
+    parser.add_argument("--workdir", metavar="PATH", help="tox working directory")
 
     parser.add_argument(
         "args", nargs="*", help="additional arguments available to command positional substitution"
     )
 
+    def _set_envdir_from_devenv(testenv_config, value):
+        if testenv_config.config.option.devenv is not None:
+            return py.path.local(testenv_config.config.option.devenv)
+        else:
+            return value
+
     parser.add_testenv_attribute(
         name="envdir",
         type="path",
         default="{toxworkdir}/{envname}",
         help="set venv directory -- be very careful when changing this as tox "
         "will remove this directory when recreating an environment",
+        postprocess=_set_envdir_from_devenv,
     )
 
     # add various core venv interpreter attributes
@@ -563,11 +547,21 @@ def tox_addoption(parser):
         python conflict is set in which case the factor name implied version if forced
         """
         for factor in testenv_config.factors:
-            if factor in tox.PYTHON.DEFAULT_FACTORS:
-                implied_python = tox.PYTHON.DEFAULT_FACTORS[factor]
+            match = tox.PYTHON.PY_FACTORS_RE.match(factor)
+            if match:
+                base_exe = {"py": "python"}.get(match.group(1), match.group(1))
+                version_s = match.group(2)
+                if not version_s:
+                    version_info = ()
+                elif len(version_s) == 1:
+                    version_info = (version_s,)
+                else:
+                    version_info = (version_s[0], version_s[1:])
+                implied_version = ".".join(version_info)
+                implied_python = "{}{}".format(base_exe, implied_version)
                 break
         else:
-            implied_python, factor = None, None
+            implied_python, version_info, implied_version = None, (), ""
 
         if testenv_config.config.ignore_basepython_conflict and implied_python is not None:
             return implied_python
@@ -575,23 +569,20 @@ def tox_addoption(parser):
         proposed_python = (implied_python or sys.executable) if value is None else str(value)
         if implied_python is not None and implied_python != proposed_python:
             testenv_config.basepython = proposed_python
-            match = tox.PYTHON.PY_FACTORS_RE.match(factor)
-            implied_version = match.group(2) if match else None
-            if implied_version is not None:
-                python_info_for_proposed = testenv_config.python_info
-                if not isinstance(python_info_for_proposed, NoInterpreterInfo):
-                    proposed_version = "".join(
-                        str(i) for i in python_info_for_proposed.version_info[0:2]
-                    )
-                    # '27'.startswith('2') or '27'.startswith('27')
-                    if not proposed_version.startswith(implied_version):
-                        # TODO(stephenfin): Raise an exception here in tox 4.0
-                        warnings.warn(
-                            "conflicting basepython version (set {}, should be {}) for env '{}';"
-                            "resolve conflict or set ignore_basepython_conflict".format(
-                                proposed_version, implied_version, testenv_config.envname
-                            )
+            python_info_for_proposed = testenv_config.python_info
+            if not isinstance(python_info_for_proposed, NoInterpreterInfo):
+                proposed_version = ".".join(
+                    str(x) for x in python_info_for_proposed.version_info[: len(version_info)]
+                )
+                if proposed_version != implied_version:
+                    # TODO(stephenfin): Raise an exception here in tox 4.0
+                    warnings.warn(
+                        "conflicting basepython version (set {}, should be {}) for env '{}';"
+                        "resolve conflict or set ignore_basepython_conflict".format(
+                            proposed_version, implied_version, testenv_config.envname
                         )
+                    )
+
         return proposed_python
 
     parser.add_testenv_attribute(
@@ -673,14 +664,17 @@ def tox_addoption(parser):
         value = list(itertools.chain.from_iterable([x.split(" ") for x in value]))
 
         passenv = {
-            "PATH",
-            "PIP_INDEX_URL",
+            "CURL_CA_BUNDLE",
             "LANG",
             "LANGUAGE",
             "LD_LIBRARY_PATH",
+            "PATH",
+            "PIP_INDEX_URL",
+            "REQUESTS_CA_BUNDLE",
+            "SSL_CERT_FILE",
             "TOX_WORK_DIR",
             str(REPORTER_TIMESTAMP_ON_ENV),
-            str(PARALLEL_ENV_VAR_KEY),
+            str(PARALLEL_ENV_VAR_KEY_PUBLIC),
         }
 
         # read in global passenv settings
@@ -783,7 +777,7 @@ def tox_addoption(parser):
 
     def develop(testenv_config, value):
         option = testenv_config.config.option
-        return not option.installpkg and (value or option.develop)
+        return not option.installpkg and (value or option.develop or option.devenv is not None)
 
     parser.add_testenv_attribute(
         name="usedevelop",
@@ -900,7 +894,7 @@ class TestenvConfig:
         #: set of factors
         self.factors = factors
         self._reader = reader
-        self.missing_subs = []
+        self._missing_subs = []
         """Holds substitutions that could not be resolved.
 
         Pre 2.8.1 missing substitutions crashed with a ConfigError although this would not be a
@@ -1040,8 +1034,8 @@ class ParseIni(object):
         config.sdistsrc = reader.getpath("sdistsrc", None)
         config.setupdir = reader.getpath("setupdir", "{toxinidir}")
         config.logdir = config.toxworkdir.join("log")
-        within_parallel = PARALLEL_ENV_VAR_KEY in os.environ
-        if not within_parallel:
+        within_parallel = PARALLEL_ENV_VAR_KEY_PRIVATE in os.environ
+        if not within_parallel and not WITHIN_PROVISION:
             ensure_empty_dir(config.logdir)
 
         # determine indexserver dictionary
@@ -1138,6 +1132,12 @@ class ParseIni(object):
 
         config.skipsdist = reader.getbool("skipsdist", all_develop)
 
+        if config.option.devenv is not None:
+            config.option.notest = True
+
+        if config.option.devenv is not None and len(config.envlist) != 1:
+            feedback("--devenv requires only a single -e", sysexit=True)
+
     def handle_provision(self, config, reader):
         requires_list = reader.getlist("requires")
         config.minversion = reader.getstring("minversion", None)
@@ -1168,19 +1168,28 @@ class ParseIni(object):
         for require in requires + [min_version]:
             # noinspection PyBroadException
             try:
-                package = pkg_resources.Requirement.parse(require)
-                if package.project_name not in exists:
+                package = requirements.Requirement(require)
+                # check if the package even applies
+                if package.marker and not package.marker.evaluate({"extra": ""}):
+                    continue
+                package_name = canonicalize_name(package.name)
+                if package_name not in exists:
                     deps.append(DepConfig(require, None))
-                    exists.add(package.project_name)
-                    pkg_resources.get_distribution(package)
-            except pkg_resources.RequirementParseError as exception:
+                    exists.add(package_name)
+                    dist = importlib_metadata.distribution(package.name)
+                    if not package.specifier.contains(dist.version, prereleases=True):
+                        raise MissingDependency(package)
+            except requirements.InvalidRequirement as exception:
                 failed_to_parse = True
                 error("failed to parse {!r}".format(exception))
             except Exception as exception:
                 verbosity1("could not satisfy requires {!r}".format(exception))
-                missing_requirements.append(str(pkg_resources.Requirement(require)))
+                missing_requirements.append(str(requirements.Requirement(require)))
         if failed_to_parse:
             raise tox.exception.BadRequirement()
+        if WITHIN_PROVISION and missing_requirements:
+            msg = "break infinite loop provisioning within {} missing {}"
+            raise tox.exception.Error(msg.format(sys.executable, missing_requirements))
         config.run_provision = bool(len(missing_requirements))
         return deps
 
@@ -1241,7 +1250,7 @@ class ParseIni(object):
                 if env_attr.postprocess:
                     res = env_attr.postprocess(testenv_config=tc, value=res)
             except tox.exception.MissingSubstitution as e:
-                tc.missing_subs.append(e.name)
+                tc._missing_subs.append(e.name)
                 res = e.FLAG
             setattr(tc, env_attr.name, res)
             if atype in ("path", "string", "basepython"):
@@ -1277,9 +1286,10 @@ class ParseIni(object):
             all_envs = self._getallenvs(reader)
         else:
             candidates = (
-                (os.environ.get(PARALLEL_ENV_VAR_KEY), True),
+                (os.environ.get(PARALLEL_ENV_VAR_KEY_PRIVATE), True),
                 (from_option, True),
                 (from_environ, True),
+                ("py" if self.config.option.devenv is not None else None, False),
                 (from_config, False),
             )
             env_str, envlist_explicit = next(((i, e) for i, e in candidates if i), ([], False))
@@ -1681,7 +1691,10 @@ class _ArgvlistReader:
     @classmethod
     def processcommand(cls, reader, command, replace=True):
         posargs = getattr(reader, "posargs", "")
-        posargs_string = list2cmdline([x for x in posargs if x])
+        if sys.platform.startswith("win"):
+            posargs_string = list2cmdline([x for x in posargs if x])
+        else:
+            posargs_string = " ".join([shlex_quote(x) for x in posargs if x])
 
         # Iterate through each word of the command substituting as
         # appropriate to construct the new command string. This
